/*

HVcMO (referred as HVcMO20a in the current setup)

By Ricardo Navarro
ricardo[at]iwate-u.ac.jp
https://github.com/ricardonrcu/HVcMO


WHAT IS HVcMO?

It is an instance of the VMO-N framework for multimodal optimization.
At the same time, HVcMO is an extension of the HillVallEA multimodal approach.


HOW DOES HVcMO MODIFIES HillVallEA?

The main augmentations concern:
 1- The application of the search operators of VMO over the truncated population (this is the biggest change).
 2- The shrinkage of the size of the population to sample when the remaining budget is not enough to deal with a new population with the
	current size. This choice was indeed considered by S.C. Maree for HillVallEA, although discarded.
 3- The alteration of the condition to increment the size parameters. In HillVallEA, that happens if no new peak is detected. Here, it occurs
	if also the population size was never reduced (this is a consequence of the 2nd modification).
 4- The inclusion of other solutions in the 'rejection set', in particular, each master that represents a recently discovered peak,
	and the best elite found thus far.


COMPATIBILITY Of THE SOURCE CODES FOR HVcMO20a AND HillVallEA19

As evolutionary algorithms, HVcMO and HillVallEA share several conceptualizations, regardless of the names given, e.g. node (in VMO) means
a solution and the mesh represents the population. In addition, there is a common reasoning behind any EA, implying that there are lots of
common procedures. Then, instead of writing a totally new source code for HVcMO20a, it largely reutilizes the source code of HillVallEA19,
the ultimate version of HillVallEA, also as an attempt to assure a fair comparability between their performances. The program for
HillVallEA19 is available online under GNU General Public License v3.0:

By S.C. Maree
https://github.com/SCMaree/HillVallEA


HOW TO RUN HVcMO20a?

 1- Clone the HillVallEA19 project from the repository told above.
 2- Add the files "hvcmo20a.h" and "hvcmo20a.cpp" (provided here) to the HillVallEA package project inside your cloned HillVallEA-master.
 3- Specify:
	  class hvcmo_t;
	  typedef std::shared_ptr<hvcmo_t> hvcmo_pt;
	in the definitions for the namespace hillvallea, on the "hillvallea_internal.hpp" file in the HillVallEA project.
 4- For a start, take the example script provided by S.C. Maree as "example_CEC2013_benchmark.cpp"
 5- Add the reference to hvcmo20a:
	  #include "HillVallEA/hvcmo20a.h"
	Later, instead of creating a hillvallea object as:
	  hillvallea::hillvallea_t hillvallea(...);
	create a hvcmo object as:
	  hillvallea::hvcmo_t hvcmo(...);
 6- Consequently set 'local_optimizer_index' and 'cluster_alg' for your hvcmo object.
 7- Invoke the execution of the optimizer by telling the id for the current run:
	  hvcmo.run(run+1);
	Such an id has no serious impact; it is used only to form some names for output files.
 8- Run the script.


LICENCE

The extensions of HVcMO with respect to HillVallEA are available under the same GNU General Public License v3.0.

*/


#include "pch.h"	// This is a precompiled header file.

#include "hvcmo20a.h"
#include "population.hpp"
#include "mathfunctions.hpp"
#include "fitness.h"
#include "hgml.hpp"

namespace hillvallea
{
	// Create nodes from those in the frontiers of the mesh.
	// Here, we distinguish between a lower frontier and an upper frontier.
	void hvcmo_t::h_operator(population_pt population, const int max_nodes_to_create, std::vector<solution_pt>& new_nodes) {
		new_nodes.clear();

		// We sort the initial mesh by norm values. Then, it is very likely that initial_mesh->sols[0] will no correspond to the
		// globally best node anymore. However, it is saved in this->best_global. Also, a reference to it is still in total_mesh->sols[0]
		population->sort_on_norm();

		int i;
		assert(max_nodes_to_create <= population->size());
		for (i = 0; i < population->size() && i < max_nodes_to_create / 2; i++) {
			new_nodes.push_back(generate_node_from_lower_frontier(population, i));
		}

		for (int p = population->size() - 1; i < max_nodes_to_create; p--, i++) {
			new_nodes.push_back(generate_node_from_upper_frontier(population, i));
		}
	}


	// Generates a node from a given node in the lower frontier of the mesh
	solution_pt hvcmo_t::generate_node_from_lower_frontier(population_pt population, const int frontier_node_index)
	{
		vec_t variable_values(this->number_of_parameters);

		for (int i = 0; i < this->number_of_parameters; i++) {
			double weight = (this->w0[i] - this->w1[i]) * (this->maximum_number_of_evaluations - this->number_of_evaluations) / (this->maximum_number_of_evaluations + this->w1[i]);
			double temp_value = 0;
			if (population->sols[frontier_node_index]->param[i] > 0)
				temp_value = std::abs(population->sols[frontier_node_index]->param[i] - weight);
			else
				temp_value = population->sols[frontier_node_index]->param[i] + weight;

			// If the value is in rage
			if (temp_value >= this->lower_param_bounds[i] && temp_value <= this->upper_param_bounds[i]) {
				variable_values[i] = temp_value;
			}
			else {
				double midpoint = (this->upper_param_bounds[i] + this->lower_param_bounds[i]) / 2;
				double min, max;

				if (temp_value < this->lower_param_bounds[i]) {
					min = this->lower_param_bounds[i];
					max = midpoint;
				}
				// Else, temp_value > this->upper_param_bounds[i]
				else {
					min = midpoint;
					max = this->upper_param_bounds[i];
				}

				std::uniform_real_distribution<double> unif_rand(min, max);
				rng->seed((unsigned long)clock());
				variable_values[i] = unif_rand(*this->rng);
			}
		}

		solution_pt new_node = std::make_shared<solution_t>(variable_values);

		this->fitness_function->evaluate(new_node);
		this->number_of_evaluations++;

		return new_node;
	}


	// Generates a node from a given node in the upper frontier of the mesh
	solution_pt hvcmo_t::generate_node_from_upper_frontier(population_pt population, const int frontier_node_index)
	{
		vec_t variable_values(this->number_of_parameters);

		for (int i = 0; i < this->number_of_parameters; i++) {
			double weight = (this->w0[i] - this->w1[i]) * (this->maximum_number_of_evaluations - this->number_of_evaluations) / (this->maximum_number_of_evaluations + this->w1[i]);
			double temp_value = 0;
			if (population->sols[frontier_node_index]->param[i] > 0)
				temp_value = population->sols[frontier_node_index]->param[i] + weight;
			else
				temp_value = std::abs(population->sols[frontier_node_index]->param[i] - weight);

			// If the value is in rage
			if (temp_value >= this->lower_param_bounds[i] && temp_value <= this->upper_param_bounds[i]) {
				variable_values[i] = temp_value;
			}
			else {
				double midpoint = (this->upper_param_bounds[i] + this->lower_param_bounds[i]) / 2;
				double min, max;

				if (temp_value < this->lower_param_bounds[i]) {
					min = this->lower_param_bounds[i];
					max = midpoint;
				}
				// Else, temp_value > this->upper_param_bounds[i]
				else {
					min = midpoint;
					max = this->upper_param_bounds[i];
				}

				rng->seed((unsigned long)clock());
				std::uniform_real_distribution<double> unif_rand(min, max);
				variable_values[i] = unif_rand(*this->rng);
			}
		}

		solution_pt new_node = std::make_shared<solution_t>(variable_values);

		this->fitness_function->evaluate(new_node);
		this->number_of_evaluations++;

		return new_node;
	}


	// Calculates the distance allowed (from any node) in the current generation (cycle).
	// This is used to define the range to create new nodes by the operator f of VMO.
	double hvcmo_t::calculate_distance_threshold()
	{
		double distance = 0;
		for (int i = 0; i < this->number_of_parameters; i++)
			distance += this->upper_param_bounds[i] - this->lower_param_bounds[i];
		distance = distance / this->number_of_parameters;

		if (this->number_of_evaluations < this->maximum_number_of_evaluations * 0.15)
			return distance / 2;
		else if (this->number_of_evaluations < this->maximum_number_of_evaluations * 0.3)
			return distance / 4;
		else if (this->number_of_evaluations < this->maximum_number_of_evaluations * 0.6)
			return distance / 8;
		else if (this->number_of_evaluations < this->maximum_number_of_evaluations * 0.8)
			return distance / 16;
		else
			return distance / 100;
	}


	// Computes the Euclidean distance between the variable vectors of two solutions
	double hvcmo_t::euclidean_distance(solution_pt& node1, solution_pt& node2)
	{
		//std::cout << node1->param.size() << " " << node2->param.size() << std::endl;
		assert(node1->param.size() == node2->param.size());
		double distance = 0;
		for (size_t i = 0; i < node1->param.size(); i++) {
			distance += (node1->param[i] - node2->param[i]) * (node1->param[i] - node2->param[i]);
		}
		return std::sqrt(distance);
	}


	// calculate the Euclidean distance between the nodes in the mesh
	void hvcmo_t::calculate_distance_matrix(population_pt population, std::vector<std::vector<double>>& distance_matrix)
	{
		// Savig the data either over or under the diagonal of the matrix would be enough.
		// However, when we later need to read the values in this matrix, corresponding to any two nodes (a 'current' node
		// and its best local/global neighbor, we need to compare their indeces every time.
		// For that reason, let's save the full matrix, excepting the diagonal.

		distance_matrix.clear();

		distance_matrix.resize(population->size());
		for (size_t i = 0; i < population->size(); i++)
			distance_matrix[i].resize(population->size());

		// Calculate and save the distance values (skip the diagonal)
		for (size_t i = 0; i < population->size(); i++) {
			for (size_t j = i + 1; j < population->size(); j++) {
				distance_matrix[i][j] = distance_matrix[j][i] = this->euclidean_distance(population->sols[i], population->sols[j]);
			}
		}
	}


	// Given the index of a node, returns the indeces of its k nearest neighbors in the subset of the expanding mesh indicated by 'num_trunc'
	void hvcmo_t::locate_neighbors_trunc(const int index, const int num_trunc, std::vector<int>& neighbors, std::vector<double>& distance_vector)
	{
		assert(neighbors.size() == this->k_number_of_neighbors);

		int k_nearest_index = index;
		double k_nearest_distance = distance_vector[index] = std::numeric_limits<double>::max();

		int k = 0;
		while (k < this->k_number_of_neighbors) {
			for (int j = 0; j < num_trunc; j++) {
				if (distance_vector[j] < k_nearest_distance) {
					k_nearest_distance = distance_vector[j];
					k_nearest_index = j;
				}
			}
			neighbors[k] = k_nearest_index;

			k_nearest_distance = distance_vector[k_nearest_index] = std::numeric_limits<double>::max();
			k++;
		}
	}


	// Given the index of a 'current' node and its neighbors (indeces in population), returns the index of its fittest neighbor (if any)
	// or returns 'index' if the current node is the fittest one (local minimum)
	int hvcmo_t::locate_local_minimum(population_pt population, const int index, std::vector<int>& neighbors)
	{
		int index_best_neighbor = index;
		double minimum_fitness = population->sols[index]->f;

		for (size_t i = 0; i < neighbors.size(); i++)
		{
			if (population->sols[neighbors[i]]->f < population->sols[index_best_neighbor]->f) {
				index_best_neighbor = neighbors[i];
				minimum_fitness = population->sols[neighbors[i]]->f;
			}

		}

		return index_best_neighbor;
	}


	// Create a node between a certain node and its fittest neighbor
	solution_pt hvcmo_t::f_operator(const solution_pt& node, const solution_pt& neighbor, const double distance_threshold) {
		vec_t variable_values(this->number_of_parameters);
		std::uniform_real_distribution<double> unif_rand(0, 1);
		rng->seed((unsigned long)clock());

		for (int i = 0; i < this->number_of_parameters; i++) {
			double midpoint = (node->param[i] + neighbor->param[i]) / 2;
			if (std::abs(midpoint - neighbor->param[i]) > distance_threshold && unif_rand(*this->rng) <= nearness_factor_minimization_case(*node, *neighbor)) {
				variable_values[i] = midpoint;
			}
			else if (std::abs(midpoint - neighbor->param[i]) <= distance_threshold) {
				// Generate a random value (val) in the range [-distance_threshold, distance_threshold]
				double val = (2 * unif_rand(*this->rng) - 1) * distance_threshold;
				variable_values[i] = neighbor->param[i] + val;
				// Boundary repair
				if (variable_values[i] > this->upper_param_bounds[i])
					variable_values[i] = this->upper_param_bounds[i];
				else if (variable_values[i] < this->lower_param_bounds[i])
					variable_values[i] = this->lower_param_bounds[i];
			}
			else {
				// Generate a random value between midpoint and the range the i-th variable value for the neighbor
				double max_val = std::max(midpoint, neighbor->param[i]);
				double min_val = std::min(midpoint, neighbor->param[i]);
				variable_values[i] = min_val + unif_rand(*this->rng) * (max_val - min_val);
			}
		}

		solution_pt new_node = std::make_shared<solution_t>(variable_values);

		this->fitness_function->evaluate(new_node);
		this->number_of_evaluations++;

		return new_node;
	}



	// Create a node between a certain node and the globally best node in the expanding mesh
	solution_pt hvcmo_t::g_operator(const solution_pt& node) {
		vec_t variable_values(this->number_of_parameters);
		std::uniform_real_distribution<double> unif_rand(0, 1);
		rng->seed((unsigned long)clock());

		for (int i = 0; i < this->number_of_parameters; i++) {
			double midpoint = (node->param[i] + this->best_in_selection.param[i]) / 2;
			if (unif_rand(*this->rng) <= nearness_factor_minimization_case(*node, this->best_in_selection)) {
				variable_values[i] = midpoint;
			}
			else {
				// Generate a random value between midpoint and the range the i-th variable value for the neighbor
				double max_val = std::max(midpoint, this->best_in_selection.param[i]);
				double min_val = std::min(midpoint, this->best_in_selection.param[i]);
				variable_values[i] = min_val + unif_rand(*this->rng) * (max_val - min_val);
			}
		}

		solution_pt new_node = std::make_shared<solution_t>(variable_values);

		this->fitness_function->evaluate(new_node);
		this->number_of_evaluations++;

		return new_node;
	}


	// Computes the nearness factor for a minimization task.
	// This is used by the operators f and g to generate nodes.
	// 'best' denotes a local/global best solution 
	double hvcmo_t::nearness_factor_minimization_case(const solution_t& node, const solution_t& best)
	{
		assert(best.f <= node.f);

		if (node.f < 0)
			return (std::abs(node.f) + 1) / (std::abs(best.f) + 1);
		else
			return (std::abs(best.f) + 1) / (node.f + std::abs(best.f) + 1);
	}


	// Constructor
	hvcmo_t::hvcmo_t(
		fitness_pt fitness_function,
		const int niching_alg_id,
		const int local_optimizer_id,
		const int number_of_parameters,
		const vec_t& lower_init_ranges,
		const vec_t& upper_init_ranges,
		const vec_t& lower_param_bounds,
		const vec_t& upper_param_bounds,
		const int maximum_number_of_evaluations,
		const int maximum_number_of_seconds,
		const double vtr,
		const bool use_vtr,
		const int random_seed,
		const bool write_generational_solutions,
		const bool write_generational_statistics,
		const std::string write_directory,
		const std::string file_appendix
	)
	{
		// copy all settings
		this->fitness_function = fitness_function;
		this->cluster_alg = niching_alg_id,
		this->local_optimizer_index = local_optimizer_id;
		this->number_of_parameters = number_of_parameters;
		this->lower_init_ranges = lower_init_ranges;
		this->upper_init_ranges = upper_init_ranges;
		this->lower_param_bounds = lower_param_bounds;
		this->upper_param_bounds = upper_param_bounds;
		this->maximum_number_of_evaluations = maximum_number_of_evaluations;
		this->maximum_number_of_seconds = maximum_number_of_seconds;
		this->vtr = vtr;
		this->use_vtr = use_vtr;
		this->random_seed = random_seed;
		this->write_generational_solutions = write_generational_solutions;
		this->write_generational_statistics = write_generational_statistics;
		this->write_directory = write_directory;
		this->file_appendix = file_appendix;

		rng = std::make_shared<std::mt19937>((unsigned long)(random_seed));
		std::uniform_real_distribution<double> unif(0, 1);

		init_default_params();

	}


	// Quick Constructor
	hvcmo_t::hvcmo_t(
		fitness_pt fitness_function,
		const int  number_of_parameters,
		const vec_t& lower_param_bounds,
		const vec_t& upper_param_bounds,
		const int maximum_number_of_evaluations,
		const int random_seed
	)
	{
		// copy all settings
		this->fitness_function = fitness_function;
		this->cluster_alg = 0; // HVC 
		this->local_optimizer_index = 1; // AMaLGaM-Univariate
		this->number_of_parameters = number_of_parameters;
		this->lower_init_ranges = lower_param_bounds; // init range == bounds
		this->upper_init_ranges = upper_param_bounds;  // init range == bounds
		this->lower_param_bounds = lower_param_bounds;
		this->upper_param_bounds = upper_param_bounds;
		this->maximum_number_of_evaluations = maximum_number_of_evaluations;
		this->maximum_number_of_seconds = 0;
		this->vtr = 0;
		this->use_vtr = false;
		this->random_seed = random_seed;
		this->write_generational_solutions = true;
		this->write_generational_statistics = true;
		this->write_directory = "D:\\res\\";
		this->file_appendix = "";
	
		rng = std::make_shared<std::mt19937>((unsigned long)(random_seed));
		std::uniform_real_distribution<double> unif(0, 1);

		init_default_params();

	}


	hvcmo_t::~hvcmo_t()
	{

	}

	void hvcmo_t::init_default_params()
	{
		// Parameters of the recursion scheme
		//---------------------------------------------
		population_size_initializer = 6.0;
		population_size_incrementer = 2.0;
		cluster_size_initializer = 0.8;
		cluster_size_incrementer = 1.1;
		add_elites_max_trials = 5;
		selection_fraction_multiplier = 1.0; // this doesn't make things better, disabled it.
		TargetTolFun = 1e-5;

		scaled_search_volume = 1.0; // reduces round-off errors and +inf errors for large number_of_parameters;
		for (size_t i = 0; i < upper_init_ranges.size(); ++i) {
			scaled_search_volume *= pow((upper_init_ranges[i] - lower_init_ranges[i]), 1.0 / number_of_parameters);
		}

		write_elitist_archive = true;
		
		clustering_max_number_of_neighbours = (size_t)(number_of_parameters + 1);
		

		this->w0.resize(this->number_of_parameters);
		this->w1.resize(this->number_of_parameters);

		for (size_t i = 0; i < this->number_of_parameters; i++)
		{
			this->w0[i] = (this->upper_param_bounds[i] - this->lower_param_bounds[i]) / 10;
			this->w1[i] = (this->upper_param_bounds[i] - this->lower_param_bounds[i]) / 100;
		}

		this->k_number_of_neighbors = 3;
		this->number_of_new_global_opts_found = 0;
		this->time_ratio = 2.0;
	}

	// Write statistic Files
	//----------------------------------------------------------------------------------
	void hvcmo_t::new_statistics_file(int run_id)
	{
		std::string filename;
		std::stringstream fname;

		fname << this->fitness_function->name() << "_run" << std::setw(2) << std::setfill('0') << run_id + 1 << "_overall_statistics" << file_appendix << ".dat";
		filename = fname.str();

		if (!this->write_directory.empty()) {
			std::stringstream dir;
			dir << "overall_statistics\\" << this->fitness_function->name() << "\\";
			
			std::string directory = this->write_directory + dir.str();

			if (std::experimental::filesystem::is_directory(directory.c_str()) || std::experimental::filesystem::create_directories(directory.c_str()))
				statistics_file.open(directory + filename, std::ofstream::out | std::ofstream::trunc);
		}
		else {
			statistics_file.open(filename, std::ofstream::out | std::ofstream::trunc);
		}

		assert(statistics_file.is_open());

		statistics_file << "pop  niche    local_gen   fes        time #elites  best_elite   ave-obj       std-obj" << std::endl;

	}

	void hvcmo_t::new_statistics_file_estimate_time(int run_id)
	{
		std::string filename;
		std::stringstream fname;

		fname << this->fitness_function->name() << "_run" << std::setw(2) << std::setfill('0') << run_id + 1 << "_stats_estimate_time" << file_appendix << ".dat";
		filename = fname.str();

		if (!this->write_directory.empty()) {
			std::stringstream dir;
			dir << "stats_estimate_time\\" << this->fitness_function->name() << "\\";

			std::string directory = this->write_directory + dir.str();

			if (std::experimental::filesystem::is_directory(directory.c_str()) || std::experimental::filesystem::create_directories(directory.c_str()))
				statistics_file_estimate_time.open(directory + filename, std::ofstream::out | std::ofstream::trunc);
		}
		else {
			statistics_file_estimate_time.open(filename, std::ofstream::out | std::ofstream::trunc);
		}

		assert(statistics_file_estimate_time.is_open());

		statistics_file_estimate_time << "pop_id  pop_size   estimated_time" << std::endl;

	}

	void hvcmo_t::write_statistics_line_population(const population_t& pop, const std::vector<optimizer_pt>& local_optimizers, const std::vector<solution_pt>& elitist_archive)
	{
		clock_t current_time = clock();
		double runtime = (double(current_time - starting_time) - file_writing_time) / CLOCKS_PER_SEC * 1000.0;

		solution_t best = *pop.first();

		for (auto sol = elitist_archive.begin(); sol != elitist_archive.end(); ++sol)
		{
			if (solution_t::better_solution(**sol, best))
			{
				best = **sol;
			}
		}

		if(number_of_generations > 0)
			statistics_file << std::endl;

		statistics_file
			<< std::setw(3) << number_of_generations + 1
			<< std::setw(9) << local_optimizers.size()
			<< std::setw(7) << 0
			<< std::setw(8) << number_of_evaluations
			<< std::setw(12) << std::fixed << std::setprecision(3) << runtime
			<< std::setw(10) << elitist_archive.size()
			<< std::setw(12) << std::scientific << std::setprecision(3) << best.f
			<< std::setw(14) << std::scientific << std::setprecision(3) << pop.average_fitness()
			<< std::setw(14) << std::scientific << std::setprecision(3) << pop.relative_fitness_std()
			<< std::endl;
	}

	void hvcmo_t::write_statistics_line_cluster(const population_t& cluster_pop, int cluster_number, int cluster_generation, const std::vector<optimizer_pt>& local_optimizers, const std::vector<solution_pt>& elitist_archive)
	{
		clock_t current_time = clock();
		double runtime = (double(current_time - starting_time) - file_writing_time) / CLOCKS_PER_SEC * 1000.0;

		solution_t best = *cluster_pop.first();

		for (auto sol = elitist_archive.begin(); sol != elitist_archive.end(); ++sol)
		{
			if (solution_t::better_solution(**sol, best))
			{
				best = **sol;
			}
		}

		statistics_file
			<< std::setw(3) << ""
			<< std::setw(9) << cluster_number + 1
			<< std::setw(7) << cluster_generation
			<< std::setw(8) << number_of_evaluations
			<< std::setw(12) << std::fixed << std::setprecision(3) << runtime
			<< std::setw(10) << elitist_archive.size()
			<< std::setw(12) << std::scientific << std::setprecision(3) << best.f
			<< std::setw(14) << std::scientific << std::setprecision(3) << cluster_pop.average_fitness()
			<< std::setw(14) << std::scientific << std::setprecision(3) << cluster_pop.relative_fitness_std()
			<< std::endl;
	}

	void hvcmo_t::write_statistics_line_estimate_time(const population_t& pop)
	{
		clock_t current_time = clock();

		// Update the estimated time
		double estimated_time = time_ratio * (double)(pop.size() * pop.size());

		std::stringstream line;
		line
			<< pop.size()
			<< " " << std::setw(8) << std::fixed << std::setprecision(3) << estimated_time;

		statistics_file_estimate_time
			<< std::setw(3) << number_of_generations + 1 << " "
			<< std::setw(6) << line.str() << std::endl;
	}


	void hvcmo_t::close_statistics_file()
	{
		statistics_file.close();
	}

	void hvcmo_t::close_statistics_file_estimate_time()
	{
		statistics_file_estimate_time.close();
	}


	// Write population
	//-------------------------------------------------------------------------------
	void hvcmo_t::write_population_file(int run_id, population_pt pop, std::vector<optimizer_pt>& local_optimizers) const
	{
		std::ofstream file;
		std::string filename;
		std::stringstream fname;
		
		fname << this->fitness_function->name() << "_run" << std::setw(2) << std::setfill('0') << run_id + 1 << "_pop" << std::setw(5) << std::setfill('0') << this->number_of_generations + 1 << file_appendix << "_sampled_pop" << ".dat";
		filename = fname.str();

		if (!this->write_directory.empty()) {
			std::stringstream dir;
			dir << "population\\" << this->fitness_function->name() << "\\run" << std::setw(2) << std::setfill('0') << run_id + 1 << "\\pop" << std::setw(5) << std::setfill('0') << this->number_of_generations + 1 << "\\";

			std::string directory = this->write_directory + dir.str();

			if (std::experimental::filesystem::is_directory(directory.c_str()) || std::experimental::filesystem::create_directories(directory.c_str()))
				file.open(directory + filename, std::ofstream::out | std::ofstream::trunc);
		}
		else {
			file.open(filename, std::ofstream::out | std::ofstream::trunc);
		}

		assert(file.is_open());

		file << "solution	=	fitness	@	penalty" << std::endl;

		for (auto sol = pop->sols.begin(); sol != pop->sols.end(); ++sol)
		{
			file << std::setprecision(std::numeric_limits<long double>::digits10 + 1) << (*sol)->param << " = "
				<< std::setprecision(std::numeric_limits<long double>::digits10 + 1) << (*sol)->f << " @ "
				<< (*sol)->penalty
				<< std::endl;
		}

		file.close();
	}

	// Write selection: initial mesh and expanded mesh
	//-------------------------------------------------------------------------------
	void hvcmo_t::write_selection_file(int run_id, population_pt pop, bool initial_mesh, std::vector<optimizer_pt>& local_optimizers) const
	{
		std::ofstream file;
		std::string filename;
		std::stringstream fname;

		if (initial_mesh)
			fname << this->fitness_function->name() << "_run" << std::setw(2) << std::setfill('0') << run_id + 1 << "_pop" << std::setw(5) << std::setfill('0') << this->number_of_generations + 1 << "_ini_mesh_(selection)" << file_appendix << ".dat";
		else
			fname << this->fitness_function->name() << "_run" << std::setw(2) << std::setfill('0') << run_id + 1 << "_pop" << std::setw(5) << std::setfill('0') << this->number_of_generations + 1 << "_exp_mesh" << file_appendix << ".dat";

		filename = fname.str();

		if (!this->write_directory.empty()) {
			std::stringstream dir;
			dir << "population\\" << this->fitness_function->name() << "\\run" << std::setw(2) << std::setfill('0') << run_id + 1 << "\\pop" << std::setw(5) << std::setfill('0') << this->number_of_generations + 1 << "\\";

			std::string directory = this->write_directory + dir.str();

			if (std::experimental::filesystem::is_directory(directory.c_str()) || std::experimental::filesystem::create_directories(directory.c_str()))
				file.open(directory + filename, std::ofstream::out | std::ofstream::trunc);
		}
		else {
			file.open(filename, std::ofstream::out | std::ofstream::trunc);
		}

		assert(file.is_open());

		file << "solution	=	fitness	@	penalty" << std::endl;

		for (auto sol = pop->sols.begin(); sol != pop->sols.end(); ++sol)
		{
			file << std::setprecision(std::numeric_limits<long double>::digits10 + 1) << (*sol)->param << " = "
				<< std::setprecision(std::numeric_limits<long double>::digits10 + 1) << (*sol)->f << " @ "
				<< (*sol)->penalty
				<< std::endl;
		}

		file.close();
	}


	// Write clusters: by the niching method (nicher) and by the local optimizer
	//-------------------------------------------------------------------------------
	void hvcmo_t::write_niche_population(int run_id, bool by_nicher, int local_generation_number, size_t cluster_number, int cluster_generation, population_pt pop) const
	{
		std::ofstream file;
		std::string filename;
		std::stringstream fname;

		if (by_nicher)
			fname << this->fitness_function->name() << "_run" << std::setw(2) << std::setfill('0') << run_id + 1 << "_pop" << std::setw(5) << std::setfill('0') << this->number_of_generations + 1 << "_niche" << std::setw(5) << std::setfill('0') << cluster_number + 1 << "_by_nicher" << file_appendix << ".dat";
		else
			fname << this->fitness_function->name() << "_run" << std::setw(2) << std::setfill('0') << run_id + 1 << "_pop" << std::setw(5) << std::setfill('0') << this->number_of_generations + 1 << "_loc_gen" << std::setw(5) << std::setfill('0') << local_generation_number + 1 << "_niche" << std::setw(5) << std::setfill('0') << cluster_number + 1 << "_by_loc_opt" << file_appendix << ".dat";

		filename = fname.str();

		if (!this->write_directory.empty()) {
			std::stringstream dir;

			if(by_nicher)
				dir << "population\\" << this->fitness_function->name() << "\\run" << std::setw(2) << std::setfill('0') << run_id + 1 << "\\pop" << std::setw(5) << std::setfill('0') << this->number_of_generations + 1 << "\\nicher\\";
			else
				dir << "population\\" << this->fitness_function->name() << "\\run" << std::setw(2) << std::setfill('0') << run_id + 1 << "\\pop" << std::setw(5) << std::setfill('0') << this->number_of_generations + 1 << "\\loc_opt\\";

			std::string directory = this->write_directory + dir.str();

			if (std::experimental::filesystem::is_directory(directory.c_str()) || std::experimental::filesystem::create_directories(directory.c_str()))
				file.open(directory + filename, std::ofstream::out | std::ofstream::trunc);
		}
		else {
			file.open(filename, std::ofstream::out | std::ofstream::trunc);
		}

		assert(file.is_open());

		file << "solution	=	fitness	@	penalty" << std::endl;

		for (auto sol = pop->sols.begin(); sol != pop->sols.end(); ++sol)
		{
			file << std::setprecision(std::numeric_limits<long double>::digits10 + 1) << (*sol)->param << " = "
				<< std::setprecision(std::numeric_limits<long double>::digits10 + 1) << (*sol)->f << " @ "
				<< (*sol)->penalty
				<< std::endl;
		}

		file.close();
	}

	void hvcmo_t::write_elitist_archive_file(int run_id, const std::vector<solution_pt>& elitist_archive) const
	{
		std::ofstream file;
		std::string filename;
		std::stringstream fname;
		
		fname << this->fitness_function->name() << "run_random_seed" << std::setw(3) << std::setfill('0') << random_seed + 1 << "_elites" << file_appendix << ".dat";
		
		filename = fname.str();

		if (!this->write_directory.empty()) {
			std::stringstream dir;
			dir << "elitist_archives\\" << this->fitness_function->name() << "\\";
						
			std::string directory = this->write_directory + dir.str();

			if (std::experimental::filesystem::is_directory(directory.c_str()) || std::experimental::filesystem::create_directories(directory.c_str()))
				file.open(directory + filename, std::ofstream::out | std::ofstream::trunc);
		}
		else {
			file.open(filename, std::ofstream::out | std::ofstream::trunc);
		}

		assert(file.is_open());

		int precision = std::numeric_limits<long double>::digits10 + 1;

		file << "solution	=	fitness	@	fes	time" << std::endl;

		for (size_t i = 0; i < elitist_archive.size(); ++i) {
			file
				<< std::fixed << std::setprecision(precision + 1) << elitist_archive[i]->param << " = "
				<< std::fixed << std::setw(precision + 5) << std::setprecision(precision + 1) << elitist_archive[i]->f << " @ "
				<< std::fixed << std::setw(precision + 5) << elitist_archive[i]->feval_obtained << " "
				<< std::fixed << std::setw(precision + 5) << std::setprecision(precision + 1) << elitist_archive[i]->time_obtained 
				<< std::endl;
		}
	}

	void hvcmo_t::write_CEC2013_niching_file(int run_id, bool final)
	{
		std::ofstream file;
		std::string filename;
		std::stringstream fname;

		if (final)
			fname << this->fitness_function->name() << "run_random_seed" << std::setw(3) << std::setfill('0') << random_seed + 1 << "_final.dat";
		else
			fname << this->fitness_function->name() << "_run" << std::setw(2) << std::setfill('0') << run_id + 1 << "_pop" << std::setw(5) << std::setfill('0') << this->number_of_generations + 1 << ".dat";

		filename = fname.str();

		if (!this->write_directory.empty()) {
			std::stringstream dir;
			dir << "elitist_archives\\" << this->fitness_function->name() << "\\";

			if (final)
				dir << "final\\";
			else
				dir << "run" << std::setw(2) << std::setfill('0') << run_id + 1 << "\\";

			std::string directory = this->write_directory + dir.str();

			if (std::experimental::filesystem::is_directory(directory.c_str()) || std::experimental::filesystem::create_directories(directory.c_str()))
				file.open(directory + filename, std::ofstream::out | std::ofstream::trunc);
		}
		else {
			file.open(filename, std::ofstream::out | std::ofstream::trunc);
		}

		assert(file.is_open());

		int precision = std::numeric_limits<long double>::digits10 + 1;

		file << "param	=	fitness	@	feval_obtained	time_obtained" << std::endl;

		for (size_t i = 0; i < elitist_archive.size(); ++i) {
			file
				<< std::fixed << std::setprecision(precision + 1) << elitist_archive[i]->param << " = "
				<< std::fixed << std::setw(precision + 5) << std::setprecision(precision + 1) << elitist_archive[i]->f << " @ "
				<< std::fixed << std::setw(precision + 5) << elitist_archive[i]->feval_obtained << " "
				<< std::fixed << std::setw(precision + 5) << std::setprecision(precision + 1) << elitist_archive[i]->time_obtained;
			file << std::endl;
		}

	}

	// Termination Criteria
	//-------------------------------------------------------------------------------
	bool hvcmo_t::terminate_on_runtime() const
	{
		// stop if we run out of time.
		if (maximum_number_of_seconds > 0)
		{
			clock_t current_time = clock();
			// We don't multiply by 1000 here (as in most cases in this project) because the time here is not intended miliseconds, but in seconds.
			double runtime = (double(current_time - starting_time) - file_writing_time) / CLOCKS_PER_SEC;

			if (runtime > maximum_number_of_seconds) {
				return true;
			}
		}

		return false;
	}

	bool hvcmo_t::terminate_on_approaching_elite(optimizer_t& local_optimizer, std::vector<solution_pt>& elite_candidates)
	{
		// find the nearest (candidate) elite that has similar or better fitness
		solution_pt nearest_elite;
		double distance_to_nearest_elite = 1e300;
		double distance;
		double best_fitness_so_far = local_optimizer.pop->sols[0]->f;
		// double TargetTolFun = 1e-5;
		if (elitist_archive.size() > 0)
		{
			// find the nearest elite
			for (size_t j = 0; j < elitist_archive.size(); ++j)
			{

				// only consider elites that have better fitness
				if (elitist_archive[j]->f < (local_optimizer.pop->sols[0]->f + TargetTolFun))
				{

					distance = elitist_archive[j]->param_distance(*local_optimizer.pop->sols[0]);
					if (distance < distance_to_nearest_elite)
					{
						distance_to_nearest_elite = distance;
						nearest_elite = elitist_archive[j];
					}

					// also find the best elite in the archive for later
					if (elitist_archive[j]->f < best_fitness_so_far) {
						best_fitness_so_far = elitist_archive[j]->f;
					}
				}

			}
		}

		for (size_t j = 0; j < elite_candidates.size(); ++j)
		{
			// only consider good elite candidates
			if (elite_candidates[j]->f < (best_fitness_so_far + TargetTolFun))
			{
				distance = elite_candidates[j]->param_distance(*local_optimizer.pop->sols[0]);

				if (distance < distance_to_nearest_elite) {
					distance_to_nearest_elite = distance;
					nearest_elite = elite_candidates[j];
				}
			}
		}

		if (nearest_elite != nullptr)
		{
			if (check_edge(*nearest_elite, *local_optimizer.pop->sols[0], 5)) {
				return true;
			}
		}

		return false;
	}

	bool hvcmo_t::terminate_on_converging_to_local_optimum(optimizer_t& local_optimizer, std::vector<solution_pt>& elite_candidates)
	{

		// if this is the first, never terminate.
		if (elite_candidates.size() == 0 && elitist_archive.size() == 0) {
			return false;
		}

		// find best solution in the archive and candidates
		//----------------------------------------------------------
		double best = 1e308;
		double max_number_of_generations_to_obtain_elite = 0;

		for (size_t i = 0; i < elitist_archive.size(); ++i) {
			if (elitist_archive[i]->f < best) {
				best = elitist_archive[i]->f;
			}

			if (elitist_archive[i]->generation_obtained > max_number_of_generations_to_obtain_elite) {
				max_number_of_generations_to_obtain_elite = elitist_archive[i]->generation_obtained;
			}
		}

		for (size_t i = 0; i < elite_candidates.size(); ++i) {
			if (elite_candidates[i]->f < best) {
				best = elite_candidates[i]->f;
			}

			if (elite_candidates[i]->generation_obtained > max_number_of_generations_to_obtain_elite) {
				max_number_of_generations_to_obtain_elite = elite_candidates[i]->generation_obtained;
			}
		}

		// Only terminate if the local optimizer if its best is significantly worse than the best elite
		if (local_optimizer.pop->sols[0]->f > best + TargetTolFun)
		{

			// compute the time to optimum
			//----------------------------------------------------------
			int lookback_window = std::min((int)local_optimizer.average_fitness_history.size(), 5);

			if (lookback_window < 5) {
				return false;
			}

			// terminate only if the recent averages are all decreasing
			// note, if this passes, it implies curr_dto < prev_dto,
			// and thus tto > 0
			for (size_t i = 0; i < lookback_window - 1; ++i)
			{
				if (local_optimizer.average_fitness_history[local_optimizer.average_fitness_history.size() - lookback_window + i]
					< local_optimizer.average_fitness_history[local_optimizer.average_fitness_history.size() - lookback_window + i + 1]) {
					return false;
				}
			}

			double curr_fitness = local_optimizer.average_fitness_history.back();
			double prev_fitness = local_optimizer.average_fitness_history[local_optimizer.average_fitness_history.size() - lookback_window];

			double curr_dto = curr_fitness - best;
			double prev_dto = prev_fitness - best;

			// this is the best so far, let it run
			if (curr_dto <= 0.0) {
				return false;
			}

			// normal case     
			double dto_vtr = 1e-12;
			double cr5 = (curr_dto - prev_dto) / prev_dto;
			double cr = cr5; // pow((1 + cr5), 1.0 / lookback_window) - 1;

			double tto = lookback_window * log(dto_vtr / curr_dto) / log(1.0 + cr);

			if ((local_optimizer.number_of_generations + tto) > 50 * max_number_of_generations_to_obtain_elite) {
				return true;
			}

		}

		return false;

	}


	void hvcmo_t::initialize(int run_id, population_pt pop, size_t population_size, double selection_fraction_multiplier, std::vector<optimizer_pt>& local_optimizers, const std::vector<solution_pt>& elitist_archive)
	{

		// Initialize running parameters of hillvallea
		//-------------------------------------------------
		local_optimizers.clear();

		// Initially, we create a single cluster that we initialize by uniform sampling
		//-------------------------------------------------------------------------------------------------------------
		std::vector<solution_pt> backup_sols = pop->sols;
		
		// Copy in the backup population the globally best solution together with the new global optima found in the prior generation.
		// First interation: best.param.size() == 0
		if (best.param.size() > 0) {
			backup_sols.push_back(std::make_shared<solution_t>(best));
			int first_index = elitist_archive.size() - number_of_new_global_opts_found;
			// If the first_index == 0 is because 'best' was discovered in the most recent generation, so we skeep it to avoid copying again.
			if (first_index == 0)
				first_index = 1;
			for (int i = first_index; i < elitist_archive.size(); i++)
				backup_sols.push_back(elitist_archive[i]);
		}

		double sample_ratio = 2.0;
		pop->fill_with_rejection(population_size, number_of_parameters, sample_ratio, backup_sols, lower_init_ranges, upper_init_ranges, rng);
		

		{
			int fevals = pop->evaluate(this->fitness_function, 0); // no elite yet.
			number_of_evaluations += fevals;
			number_of_evaluations_init += fevals;
		}

		pop->sort_on_fitness();

		// create a dummy local_optimizer for the initial population so that we can perform selection and we can write it down.
		double init_univariate_bandwidth = scaled_search_volume * pow(pop->size(), -1.0 / number_of_parameters);
		optimizer_pt local_optimizer = init_optimizer(local_optimizer_index, number_of_parameters, lower_param_bounds, upper_param_bounds, init_univariate_bandwidth, fitness_function, rng);
		local_optimizer->initialize_from_population(pop);

		population_pt selection = std::make_shared<population_t>();
		local_optimizer->pop->truncation_percentage(*selection, local_optimizer->selection_fraction * selection_fraction_multiplier);
		

		if (write_generational_solutions)
		{
			clock_t write_starting_time = clock();

			#pragma omp critical (output_block)
			{
				write_selection_file(run_id, selection, true, local_optimizers);		// initial mesh (selection)
			}
			
			file_writing_time += double(clock() - write_starting_time);
		}

		population_pt expanding_mesh = std::make_shared<population_t>();

		// Operator f: generates nodes towards the locally best nodes (in a neighborhood).
		// A neighborhood consists of the current (i-th) node and its k nearest neighbors.
		// Because initial_mesh->sols[0] means the best node in 'selection', we skip i = 0 here:
		std::vector<std::vector<double>> distance_matrix;
		this->calculate_distance_matrix(selection, distance_matrix);
		int num_trunc = std::ceil(selection->size() * 0.35);

		for (size_t i = 1; i < selection->size(); i++) {
			// Find the neighbors of the current node (i)
			std::vector<int> neighbors(this->k_number_of_neighbors);
			locate_neighbors_trunc(i, num_trunc, neighbors, distance_matrix[i]);

			int local_minimum = locate_local_minimum(selection, i, neighbors);
			int distance_allowed = calculate_distance_threshold();

			// If the current node is not the local minimum the analized neighborhood
			if (local_minimum != i) {
				// Create a new node between the current node and the local minimum, and add it to the total mesh
				solution_pt new_node = f_operator(selection->sols[i], selection->sols[local_minimum], distance_allowed);
				expanding_mesh->sols.push_back(new_node);
			}
		}

		best_in_selection = *selection->sols[0];

		for (size_t i = 1; i < selection->size(); i++) {
			// Create a new node between the current node and the global best, and add it to the total mesh
			solution_pt new_node = g_operator(selection->sols[i]);
			expanding_mesh->sols.push_back(new_node);
		}

		// Operator h: generates nodes from those in the lower and upper frontiers of the mesh.
		// Calculate the number of nodes to create in order to complete the mesh expansion.
		size_t max_nodes_to_create = (size_t)std::min((int)std::ceil(selection->size() * 0.05), 50);
		// Create the new nodes
		std::vector<solution_pt> new_nodes;
		h_operator(selection, max_nodes_to_create, new_nodes);
		// Add the new nodes to the total mesh
		for (size_t i = 0; i < new_nodes.size(); i++) {
			expanding_mesh->sols.push_back(new_nodes[i]);
		}

		for (size_t i = 0; i < expanding_mesh->size(); i++)
			selection->sols.push_back(expanding_mesh->sols[i]);

		selection->sort_on_fitness();
				

		// Hill-Valley Clustering
		// note: as the univariate bandwidth initializes based on the average edge length, this is correct here
		//------------------------------------------------
		std::vector<population_pt> clusters;

		if (cluster_alg == 1)
		{
			hgml_t hgml;
			hgml.hierarchical_clustering(*selection, clusters);
		}
		else
			if (cluster_alg == 2)
			{
				hgml_t hgml;
				hgml.nearest_better_clustering(*selection, clusters);
			}
			else
			{
				// add elites to the selection, and mark them as elite
				if (elitist_archive.size() > 0)
				{
					for (size_t i = 0; i < elitist_archive.size(); ++i)
					{
						elitist_archive[i]->elite = true;
						selection->sols.push_back(std::make_shared<solution_t>(*elitist_archive[i]));
					}
					selection->sort_on_fitness();
				}
				

				hillvalley_clustering(*selection, clusters);

				// If unexpectedly, no clusters are found, the whole population (selection) is naturally assumed as the unique cluster.
				if (clusters.size() == 0) {
					clusters.push_back(selection);
				}
			}

		// Init local optimizers
		//---------------------------------------------------------------------------
		init_univariate_bandwidth = scaled_search_volume * pow(selection->size(), -1.0 / number_of_parameters);
		for (auto cluster = clusters.begin(); cluster != clusters.end(); ++cluster)
		{
			if ((*cluster)->sols.size() > 0)
			{
				optimizer_pt opt = init_optimizer(local_optimizer_index, number_of_parameters, lower_param_bounds, upper_param_bounds, init_univariate_bandwidth, fitness_function, rng);
				opt->initialize_from_population(*cluster);
				opt->average_fitness_history.push_back(opt->pop->average_fitness());
				local_optimizers.push_back(opt);
			}
		}


		if (write_generational_statistics) {
			clock_t write_starting_time = clock();
			
			#pragma omp critical (output_block)
			{
				write_statistics_line_population(*pop, local_optimizers, elitist_archive);
			}

			file_writing_time += double(clock() - write_starting_time);
		}
		
		if (write_generational_solutions)
		{
			clock_t write_starting_time = clock();

			#pragma omp critical (output_block)
			{
				write_population_file(run_id, pop, local_optimizers);				// sampled population
				write_selection_file(run_id, selection, false, local_optimizers);	// expanded mesh

				for (size_t i = 0; i < local_optimizers.size(); ++i)
					write_niche_population(run_id, true, number_of_generations, i, local_optimizers[i]->number_of_generations, local_optimizers[i]->pop);
			}

			file_writing_time += double(clock() - write_starting_time);
		}

	}


	void hvcmo_t::run(int run_id)
	{

		// ugly hack to prevent CPU burning in CEC problems
		// this is only used to terminate when the elitist archive
		// contains this many solutions. We dot not actually check
		// whether the solutions are distinct global optima
		// (but they will be, trust me)
		//---------------------------------------------------
		vec_t number_of_optima(20, 0.0);
		number_of_optima[0] = 2;
		number_of_optima[1] = 5;
		number_of_optima[2] = 1;
		number_of_optima[3] = 4;
		number_of_optima[4] = 2;
		number_of_optima[5] = 18;
		number_of_optima[6] = 36;
		number_of_optima[7] = 81;
		number_of_optima[8] = 216;
		number_of_optima[9] = 12;
		number_of_optima[10] = 6;
		number_of_optima[11] = 8;
		number_of_optima[12] = 6;
		number_of_optima[13] = 6;
		number_of_optima[14] = 8;
		number_of_optima[15] = 6;
		number_of_optima[16] = 8;
		number_of_optima[17] = 6;
		number_of_optima[18] = 8;
		number_of_optima[19] = 8;



		//---------------------------------------------
		// reset all runlogs (in case the multimodal optimizer is run multiple time)
		starting_time = clock();
		file_writing_time = 0;
		success = false;
		terminated = false;
		number_of_evaluations = 0;
		number_of_evaluations_init = 0;
		number_of_evaluations_clustering = 0;
		number_of_generations = 0;
		bool restart = true;
		int number_of_generations_without_new_clusters = 0;
		elitist_archive.clear();
				
		// allocate population
		//---------------------------------------------
		pop = std::make_shared<population_t>();

		// Init population sizes
		//---------------------------------------------
		double current_population_size = pow(2.0, population_size_initializer);
		double basic_population_size = current_population_size;

		double current_cluster_size;
		double current_selection_fraction_multiplier = 1.0;
				
		{
			hillvallea::optimizer_pt dummy_optimizer = init_optimizer(local_optimizer_index, number_of_parameters, lower_param_bounds, upper_param_bounds, 1.0, fitness_function, rng);
			current_cluster_size = cluster_size_initializer * dummy_optimizer->recommended_popsize(number_of_parameters);
		}

		if (write_generational_statistics) {
			clock_t write_starting_time = clock();

			#pragma omp critical (output_block)
			{
				new_statistics_file(run_id);
				new_statistics_file_estimate_time(run_id);
			}

			file_writing_time += double(clock() - write_starting_time);
		}

		// If required, increase the population size only if it was never reduced
		bool do_increase = true;
		
		initial_mesh.clear();

		// The big restart scheme
		//--------------------------------------------

		while (restart)
		{
		  // each restart, collect the elite_candidates and add them to the archive at the end of the run
		  // directly adding them to the archive is more expensive in terms of fevals, as you might find a bunch of local optima first
		  // and adding solutions to the elitist archive costs fevals because we check them using the HillVallyTest
			std::vector<solution_pt> elite_candidates;
			std::vector<optimizer_pt> local_optimizers;

			// If the init pop size is too large, reduce it if the remaining budget is enough to expand at least a few nodes
			bool try_to_decrease_initial_mesh = false;
			double minimum_expected_evaluations = current_population_size * 2.0 + 50;
			
			if (this->maximum_number_of_evaluations > 0 && minimum_expected_evaluations > (this->maximum_number_of_evaluations - this->number_of_evaluations)) {
				try_to_decrease_initial_mesh = true;
				do_increase = false;
			}

			while (try_to_decrease_initial_mesh)
			{
				current_population_size /= 3.0;

				minimum_expected_evaluations = current_population_size * 2.0 + 50;

				if (minimum_expected_evaluations <= (this->maximum_number_of_evaluations - this->number_of_evaluations) || current_population_size < basic_population_size) {
					try_to_decrease_initial_mesh = false;
				}
			}

			if (current_population_size < basic_population_size) {
				restart = false;	// init popsize too large
				break;
			}


			for (size_t problem = 1; problem <= 20; problem++)
			{
				std::stringstream ss;
				ss << "CEC2013_p" << std::setw(2) << std::setfill('0') << problem;
				if (fitness_function->name().compare(ss.str()) == 0)
				{
					if (elitist_archive.size() == number_of_optima[problem - 1]) {
						restart = false;
						break;
					}
				}

				if (!restart) {
					break;
				}
			}

			// compute initial population
			initialize(run_id, pop, (size_t)current_population_size, current_selection_fraction_multiplier, local_optimizers, elitist_archive);
			
			// we only create local optimizers from the global opts
			// so the local optimizer still inits new global opts
			// therefore, this is basically never hit.
			if (local_optimizers.size() == 0)
			{
				current_population_size *= population_size_incrementer * population_size_incrementer;
				current_selection_fraction_multiplier *= selection_fraction_multiplier;
				number_of_generations_without_new_clusters++;

				if (number_of_generations_without_new_clusters == 3) {
					restart = false;
					std::cout << "unfinished. ";
					break;
				}
				else {
					continue;
				}
			}
			else {
				number_of_generations_without_new_clusters = 0;
			}

			// Run each of the local optimizers until convergence
			for (size_t i = 0; i < local_optimizers.size(); ++i)
			{
				// current local optimizer
				while (true)
				{
					// stop if the feval budget is reached
					size_t fevals_needed_to_check_elites = 1 + (size_t)(elite_candidates.size() * add_elites_max_trials * (elitist_archive.size() + elite_candidates.size() * 0.5));

					if (maximum_number_of_evaluations > 0 && number_of_evaluations + fevals_needed_to_check_elites + current_cluster_size >= maximum_number_of_evaluations) {
						restart = false;
						if (local_optimizers[i]->pop->size() > 0) {
							elite_candidates.push_back(local_optimizers[i]->pop->sols[0]);
						}
						break;
					}

					// stop if we run out of time.
					if (terminate_on_runtime()) {
						restart = false;
						if (local_optimizers[i]->pop->size() > 0) {
							elite_candidates.push_back(local_optimizers[i]->pop->sols[0]);
						}
						break;
					}

					// stop if the vtr is hit
					if (use_vtr && local_optimizers[i]->pop->sols[0]->f < vtr)
					{
						restart = false;
						best = *local_optimizers[i]->pop->sols[0];
						success = true;
						break;
					}

					// stop this local optimizer if it approaches a previously obtained elite (candidate) 
					if ((1 + local_optimizers[i]->number_of_generations) % 5 == 0)
					{
						if (terminate_on_approaching_elite(*local_optimizers[i], elite_candidates)) {
							local_optimizers[i]->active = false;
							break;
						}
					}

					// stop this local optimizer if it converges to a local optimum
					if (terminate_on_converging_to_local_optimum(*local_optimizers[i], elite_candidates)) {
						local_optimizers[i]->active = false;
						break;
					}

					// if the cluster is active, and after checking it, it is terminated, 
					// we add the best solution to the elitist archive
					if (local_optimizers[i]->active && local_optimizers[i]->checkTerminationCondition())
					{
						if (local_optimizers[i]->pop->size() > 0)
						{
							if (elitist_archive.size() == 0 || local_optimizers[i]->pop->sols[0]->f < elitist_archive[0]->f + TargetTolFun) {
								elite_candidates.push_back(local_optimizers[i]->pop->sols[0]);
								elite_candidates.back()->generation_obtained = local_optimizers[i]->number_of_generations;
							}

							break;
						}
					}

					// if it is still active, run a generation of the local optimizer
					if (local_optimizers[i]->active)
					{

						local_optimizers[i]->estimate_sample_parameters();

						int local_number_of_evaluations = (int)local_optimizers[i]->sample_new_population((size_t)current_cluster_size);
						number_of_evaluations += local_number_of_evaluations;

						if (write_generational_solutions) {
							clock_t write_starting_time = clock();

							#pragma omp critical (output_block)
							{
								write_niche_population(run_id, false, number_of_generations, i, local_optimizers[i]->number_of_generations, local_optimizers[i]->pop);
							}
							
							file_writing_time += double(clock() - write_starting_time);
						}

						local_optimizers[i]->pop->truncation_percentage(*local_optimizers[i]->pop, local_optimizers[i]->selection_fraction);
						local_optimizers[i]->average_fitness_history.push_back(local_optimizers[i]->pop->average_fitness());

						if (write_generational_statistics) {
							clock_t write_starting_time = clock();

							#pragma omp critical (output_block)
							{
								write_statistics_line_cluster(*local_optimizers[i]->pop, (int)i, local_optimizers[i]->number_of_generations, local_optimizers, elitist_archive);
							}
							
							file_writing_time += double(clock() - write_starting_time);
						}
					}
				}
			}


			// check if the elites are novel and add the to the archive.
			number_of_new_global_opts_found = 0;
			int number_of_global_opts_found = 0;
			add_elites_to_archive(elitist_archive, elite_candidates, number_of_global_opts_found, number_of_new_global_opts_found);
			
			if (number_of_new_global_opts_found == 0 && do_increase) {
				current_cluster_size *= cluster_size_incrementer;
				current_population_size *= population_size_incrementer;
			}

			if (write_generational_statistics) {
				clock_t write_starting_time = clock();

				#pragma omp critical (output_block)
				{
					write_statistics_line_estimate_time(*pop);
				}
				
				file_writing_time += double(clock() - write_starting_time);
			}

			number_of_generations++;

		}

		// sort the archive s.t. the best is first
		if (elitist_archive.size() > 0) {
			best = *elitist_archive[0];
		}

		// write the final solution(s)
		// only if we care to output anything. Else, write nothing 
		if ((write_generational_statistics || write_generational_solutions) && write_elitist_archive) {
			clock_t write_starting_time = clock();

#pragma omp critical (output_block)
				{
					write_elitist_archive_file(run_id, elitist_archive);
				}

			file_writing_time += double(clock() - write_starting_time);
		}

		if (write_generational_statistics) {
			clock_t write_starting_time = clock();

			#pragma omp critical (output_block)
			{
				close_statistics_file();
				close_statistics_file_estimate_time();
			}
			
			file_writing_time += double(clock() - write_starting_time);
		}

	}


	void hvcmo_t::add_elites_to_archive(std::vector<solution_pt>& elitist_archive, const std::vector<solution_pt>& elite_candidates, int& number_of_global_opts_found, int& number_of_new_global_opts_found)
	{

		number_of_global_opts_found = 0;
		number_of_new_global_opts_found = 0;

		// find best solution in the archive
		double best_archive = 1e308;
		for (size_t i = 0; i < elitist_archive.size(); ++i)
		{
			if (elitist_archive[i]->f < best_archive) {
				best_archive = elitist_archive[i]->f;
			}
		}

		// find best candidate
		double best_candidate = 1e308;
		for (size_t i = 0; i < elite_candidates.size(); ++i)
		{
			if (elite_candidates[i]->f < best_candidate) {
				best_candidate = elite_candidates[i]->f;
			}
		}

		// clear the archive if the best candidate is better
		// than the best in the archive
		double best = std::min(best_candidate, best_archive);
		if ((best_candidate + TargetTolFun) < best_archive) {
			elitist_archive.clear();
		}

		// potential candidates are only those that are global optima 
		std::vector<solution_pt> potential_candidates;
		for (size_t i = 0; i < elite_candidates.size(); ++i)
		{
			if (elite_candidates[i]->f < (best + TargetTolFun)) {
				potential_candidates.push_back(elite_candidates[i]);
			}
		}

		// for each candidate, check if it is a novel local optimum
		for (size_t i = 0; i < potential_candidates.size(); ++i)
		{

			// check if the potential global optima is novel
			bool novel = true;

			if (elitist_archive.size() > 0)
			{
				size_t nearest_elite = 0;
				double nearest_dist = 1e300;
				double current_dist = 0;

				for (size_t j = 0; j < elitist_archive.size(); ++j)
				{

					current_dist = elitist_archive[j]->param_distance(*potential_candidates[i]);

					if (current_dist < nearest_dist)
					{
						nearest_dist = current_dist;
						nearest_elite = j;
					}
				}

				size_t j = nearest_elite; // lazy : re-using code..

				// a valid edge (check_edge returns true) suggest that the two optima are the same.
				if (check_edge(*elitist_archive[j], *potential_candidates[i], add_elites_max_trials))
				{
					novel = false;
					number_of_global_opts_found++;

					// replace the elite with the candidate if it is better
					if (solution_t::better_solution_via_pointers(potential_candidates[i], elitist_archive[j])) {
						elitist_archive[j] = std::make_shared<solution_t>(*potential_candidates[i]);
						elitist_archive[j]->elite = true;
						elitist_archive[j]->time_obtained = (double(clock() - starting_time) - file_writing_time) / CLOCKS_PER_SEC * 1000.0;
						elitist_archive[j]->feval_obtained = number_of_evaluations;
						elitist_archive[j]->generation_obtained = potential_candidates[i]->generation_obtained;
					}

				}
			}

			// it is novel, add it to the archive
			if (novel) {
				elitist_archive.push_back(std::make_shared<solution_t>(*potential_candidates[i]));
				elitist_archive.back()->elite = true;
				elitist_archive.back()->time_obtained = (double(clock() - starting_time) - file_writing_time) / CLOCKS_PER_SEC * 1000.0;
				elitist_archive.back()->feval_obtained = number_of_evaluations;
				elitist_archive.back()->generation_obtained = potential_candidates[i]->generation_obtained;
				number_of_new_global_opts_found++;
			}

		}

	}

}



// returns true if it is a valid edge. (if the solutions belong to the same basin)
bool hillvallea::hvcmo_t::check_edge(const hillvallea::solution_t& sol1, const hillvallea::solution_t& sol2, int max_trials)
{
	std::vector<solution_pt> test_points;
	return check_edge(sol1, sol2, max_trials, test_points);
}

bool hillvallea::hvcmo_t::check_edge(const hillvallea::solution_t& sol1, const hillvallea::solution_t& sol2, int max_trials, std::vector<solution_pt>& test_points)
{

	if (sol1.param_distance(sol2) == 0) {
		return true;
	}

	// elites are already checked to be in different basins
	if (sol1.elite && sol2.elite) {
		return false;
	}

	// check max_trials with number_of_evaluations remaining
	if (maximum_number_of_evaluations > 0 && max_trials > maximum_number_of_evaluations - number_of_evaluations) {
		max_trials = maximum_number_of_evaluations - number_of_evaluations;
	}

	// find the worst solution of the two. 
	solution_t worst;
	if (solution_t::better_solution(sol1, sol2)) {
		worst = sol2;
	}
	else {
		worst = sol1;
	}

	for (size_t k = 0; k < max_trials; k++)
	{

		solution_pt x_test = std::make_shared<solution_t>(sol1.param.size());

		x_test->param = sol1.param + ((k + 1.0) / (max_trials + 1.0)) * (sol2.param - sol1.param);

		fitness_function->evaluate(x_test);
		number_of_evaluations++;
		number_of_evaluations_clustering++;

		test_points.push_back(x_test);

		// if f[i] is better than f_test, we don't like the connection. So we stop.
		if (solution_t::better_solution(worst, *x_test)) {
			return false;
		}
	}

	return true;

}

void hillvallea::hvcmo_t::hillvalley_clustering(population_t& pop, std::vector<population_pt>& clusters)
{

	// reset all clusters
	clusters.clear();

	// exclude the trivial case
	if (pop.size() == 0) {
		return;
	}


	// Initialize the first cluster as the cluster containing the best (=first) solution
	//-------------------------------------------------------------------------------------
	size_t number_of_clusters = 1;
	std::vector<size_t> cluster_index(pop.size(), -1);
	cluster_index[0] = 0;

	// remember how many solutions in the population created, so that we can allocate them later. 
	std::vector<solution_pt> test_points;
	std::vector<size_t> cluster_index_of_test_points;
	double average_edge_length = scaled_search_volume * pow(pop.size(), -1.0 / number_of_parameters);
	double* dist = (double*)Malloc((long)pop.size() * sizeof(double));

	for (size_t i = 1; i < pop.size(); i++)
	{

		// compute the distance to all better solutions. 
		dist[i] = 0.0;
		size_t nearest_better_index = 0, worst_better_index = 0;
		for (size_t j = 0; j < i; j++) {
			dist[j] = pop.sols[i]->param_distance(*pop.sols[j]);

			if (dist[j] < dist[nearest_better_index]) {
				nearest_better_index = j;
			}

			if (dist[j] > dist[worst_better_index]) {
				worst_better_index = j;
			}
		}

		// Check neighbours
		bool edge_added = false;
		std::vector<size_t> does_not_belong_to(clustering_max_number_of_neighbours, -1);
		size_t old_nearest_better_index;
		std::vector<solution_pt> new_test_points_for_this_sol;

		for (size_t j = 0; j < std::min(i, clustering_max_number_of_neighbours); j++)
		{

			// find the next-to nearest index
			if (j > 0)
			{
				old_nearest_better_index = nearest_better_index;
				nearest_better_index = worst_better_index;

				for (size_t k = 0; k < i; k++) {

					if (dist[k] > dist[old_nearest_better_index] && dist[k] < dist[nearest_better_index]) {
						nearest_better_index = k;
					}
				}
			}

			bool skip_neighbour = false;
			for (size_t k = 0; k < does_not_belong_to.size(); ++k)
			{
				if (does_not_belong_to[k] == cluster_index[nearest_better_index])
				{
					skip_neighbour = true;
					break;
				}
			}

			if (skip_neighbour) {
				continue;
			}

			int max_number_of_trial_solutions = 1 + (int)(dist[nearest_better_index] / average_edge_length);
			std::vector<solution_pt> new_test_points;
			bool force_accept = false;

			if (i > 0.5 * pop.size() && max_number_of_trial_solutions == 1) {
				force_accept = true;
			}

			if (force_accept || check_edge(*pop.sols[i], *pop.sols[nearest_better_index], max_number_of_trial_solutions, new_test_points))
			{
				cluster_index[i] = cluster_index[nearest_better_index];
				edge_added = true;

				// if the edge is accepted, add all test_poitns to their cluster
				for (size_t k = 0; k < new_test_points.size(); ++k) {
					test_points.push_back(new_test_points[k]);
					cluster_index_of_test_points.push_back(cluster_index[nearest_better_index]);
				}

				break;
			}
			else
			{
				does_not_belong_to[j] = cluster_index[nearest_better_index];

				// if the edge is not accepted, add all solutions to that cluster
				// all but the last because that one caused the rejection
				if (new_test_points.size() > 0) {
					for (size_t k = 0; k < new_test_points.size() - 1; ++k) {
						new_test_points_for_this_sol.push_back(new_test_points[k]);
					}
				}

			}
		}

		// its a new clusters, label it like that. 
		if (!edge_added)
		{
			cluster_index[i] = number_of_clusters;
			number_of_clusters++;

			// if its a new cluster, add all its testpoints as well. 
			for (size_t k = 0; k < new_test_points_for_this_sol.size(); ++k)
			{
				test_points.push_back(new_test_points_for_this_sol[k]);
				cluster_index_of_test_points.push_back(cluster_index[i]);
			}

		}

	}

	// create & fill the clusters
	//---------------------------------------------------------------------------
	std::vector<population_pt> candidate_clusters(number_of_clusters);
	std::vector<bool> cluster_active(number_of_clusters, true);

	for (size_t i = 0; i < number_of_clusters; ++i) {
		candidate_clusters[i] = std::make_shared<population_t>();
	}

	for (size_t i = 0; i < cluster_index.size(); ++i) {
		candidate_clusters[cluster_index[i]]->sols.push_back(pop.sols[i]);
		pop.sols[i]->cluster_number = (int)cluster_index[i];

		// only if the elite is the best of that population, we do not run it again.
		// don't get confused by this. We disable the cluster as soon as the first solution is an elite. 
		if (candidate_clusters[cluster_index[i]]->sols.size() == 1 && pop.sols[i]->elite) {
			cluster_active[cluster_index[i]] = false;
		}
	}
	   	 

	for (size_t i = 0; i < test_points.size(); ++i) {
		candidate_clusters[cluster_index_of_test_points[i]]->sols.push_back(test_points[i]);
	}

	for (size_t i = 0; i < candidate_clusters.size(); ++i)
	{
		if (cluster_active[i]) {
			clusters.push_back(candidate_clusters[i]);
		}
	}

}
