/*

HVcMO (referred as HVcMO20a in the current setup)

By Ricardo Navarro
ricardo[at]iwate-u.ac.jp
https://github.com/ricardonrcu/HVcMO


WHAT IS HVcMO?

It is an instance of the VMO-N framework for multimodal optimization.
At the same time, HVcMO is an extension of the HillVallEA multimodal approach.


HOW DOES HVcMO MODIFIES HillVallEA?

The main augmentations concern:
 1- The application of the search operators of VMO over the truncated population (this is the biggest change).
 2- The shrinkage of the size of the population to sample when the remaining budget is not enough to deal with a new population with the
	current size. This choice was indeed considered by S.C. Maree for HillVallEA, although discarded.
 3- The alteration of the condition to increment the size parameters. In HillVallEA, that happens if no new peak is detected. Here, it occurs
	if also the population size was never reduced (this is a consequence of the 2nd modification).
 4- The inclusion of other solutions in the 'rejection set', in particular, each master that represents a recently discovered peak,
	and the best elite found thus far.


COMPATIBILITY Of THE SOURCE CODES FOR HVcMO20a AND HillVallEA19

As evolutionary algorithms, HVcMO and HillVallEA share several conceptualizations, regardless of the names given, e.g. node (in VMO) means
a solution and the mesh represents the population. In addition, there is a common reasoning behind any EA, implying that there are lots of
common procedures. Then, instead of writing a totally new source code for HVcMO20a, it largely reutilizes the source code of HillVallEA19,
the ultimate version of HillVallEA, also as an attempt to assure a fair comparability between their performances. The program for
HillVallEA19 is available online under GNU General Public License v3.0:

By S.C. Maree
https://github.com/SCMaree/HillVallEA


HOW TO RUN HVcMO20a?

 1- Clone the HillVallEA19 project from the repository told above.
 2- Add the files "hvcmo20a.h" and "hvcmo20a.cpp" (provided here) to the HillVallEA package project inside your cloned HillVallEA-master.
 3- Specify:
	  class hvcmo_t;
	  typedef std::shared_ptr<hvcmo_t> hvcmo_pt;
	in the definitions for the namespace hillvallea, on the "hillvallea_internal.hpp" file in the HillVallEA project.
 4- For a start, take the example script provided by S.C. Maree as "example_CEC2013_benchmark.cpp"
 5- Add the reference to hvcmo20a:
	  #include "HillVallEA/hvcmo20a.h"
	Later, instead of creating a hillvallea object as:
	  hillvallea::hillvallea_t hillvallea(...);
	create a hvcmo object as:
	  hillvallea::hvcmo_t hvcmo(...);
 6- Consequently set 'local_optimizer_index' and 'cluster_alg' for your hvcmo object.
 7- Invoke the execution of the optimizer by telling the id for the current run:
	  hvcmo.run(run+1);
	Such an id has no serious impact; it is used only to form some names for output files.
 8- Run the script.


LICENCE

The extensions of HVcMO with respect to HillVallEA are available under the same GNU General Public License v3.0.

*/


#pragma once

#include "hillvallea_internal.hpp"
#include "optimizer.hpp"
#include <experimental/filesystem> // create directories
#include <array>
#include <filesystem>



namespace hillvallea
{
  
  class hvcmo_t
  {
    
    
  public:
 
	  int k_number_of_neighbors;
	  vec_t w0;	// initial factor to consider during the generation of nodes from those in the boundaries of the mesh
	  vec_t w1;	// final factor
	  int number_of_new_global_opts_found;
	  int total_mesh_size;
	  std::vector<solution_pt> initial_mesh;
	  
	  // VMO operators 
	  solution_pt f_operator(const solution_pt& node, const solution_pt& neighbor, const double max_distance_allowed);
	  solution_pt g_operator(const solution_pt& node);
	  void h_operator(population_pt population, const int max_nodes_to_create, std::vector<solution_pt>& new_nodes);

	  double nearness_factor_minimization_case(const solution_t& node, const solution_t& best);
	  void locate_neighbors_trunc(const int index, const int num_trunc, std::vector<int>& neighbors, std::vector<double>& distance_vector);
	  
	  int locate_local_minimum(population_pt population, const int index, std::vector<int>& neighbors);
	  void calculate_distance_matrix(population_pt population, std::vector<std::vector<double>>& distance_matrix);
	  double euclidean_distance(solution_pt& node1, solution_pt& node2);
	  double calculate_distance_threshold();
	  solution_pt generate_node_from_upper_frontier(population_pt population, const int frontier_node_index);
	  solution_pt generate_node_from_lower_frontier(population_pt population, const int frontier_node_index);
	  
	  
	  // Constructor
	  hvcmo_t(
      fitness_pt fitness_function,
	  const int niching_alg_id,
	  const int local_optimizer_id,
      const int  number_of_parameters,
      const vec_t & lower_init_ranges,
      const vec_t & upper_init_ranges,
      const vec_t & lower_param_bounds,
      const vec_t & upper_param_bounds,
      const int maximum_number_of_evaluations,
      const int maximum_number_of_seconds,
      const double vtr,
      const bool use_vtr,
      const int random_seed,
      const bool write_generational_solutions,
      const bool write_generational_statistics,
      const std::string write_directory,
      const std::string file_appendix
    );

    // quick constructor
	  hvcmo_t(
      fitness_pt fitness_function,
      const int  number_of_parameters,
      const vec_t & lower_param_bounds,
      const vec_t & upper_param_bounds,
      const int maximum_number_of_evaluations,
      const int random_seed
    );

    ~hvcmo_t();
    
    fitness_pt fitness_function;
    int local_optimizer_index;
    int cluster_alg;
    void init_default_params();

    //--------------------------------------------------------------------------------
    // Runs minimizer
    void run(int run_id);

    // data members : optimization results
    //--------------------------------------------------------------------------------
    solution_t best;
	solution_t best_in_selection;
    std::vector<solution_pt> elitist_archive;
    bool terminated;
    bool success;
    int number_of_evaluations;
    int number_of_evaluations_init;
    int number_of_evaluations_clustering;
    int number_of_generations;
    double selection_fraction_multiplier;
	clock_t starting_time;
	double file_writing_time;	// This is the time wasted to write the output files
	double time_ratio;			// time_ratio = time_HVcMO(n) / time_HillVallEA19(n)	(estimated as 2.0)


    // Algorithm Parameters: initialized by their default values
    //-------------------------------------------------------------------------------
    double population_size_initializer;
    double population_size_incrementer;
    double cluster_size_initializer;
    double cluster_size_incrementer;
    double scaled_search_volume;
    size_t clustering_max_number_of_neighbours;
    double TargetTolFun;
    int add_elites_max_trials;

    // Hill-Valley Test and Clustering
    //--------------------------------------------------------------------------------
    void hillvalley_clustering(population_t & pop, std::vector<population_pt> & clusters);
    bool check_edge(const solution_t & sol1, const solution_t & sol2, int max_trials);
    bool check_edge(const solution_t & sol1, const solution_t & sol2, int max_trials, std::vector<solution_pt> & test_points);

    // Random number generator
    // Mersenne twister
    //------------------------------------
    std::shared_ptr<std::mt19937> rng;
    bool write_elitist_archive;
    bool write_generational_solutions;
    bool write_generational_statistics;
	std::string write_directory;
    
  private:

    // data members : user settings
    //-------------------------------------------------------------------------------
    int  number_of_parameters;
    vec_t  lower_init_ranges;
    vec_t  upper_init_ranges;
    vec_t  lower_param_bounds;
    vec_t  upper_param_bounds;
    int maximum_number_of_evaluations;
    int maximum_number_of_seconds;
    double vtr;
    bool use_vtr;
    int random_seed;
    std::string file_appendix;
    std::string file_appendix_generation;


    // Run-time functions
    //-------------------------------------------------------------------------------
	void initialize(int run_id, population_pt pop, size_t population_size, double selection_fraction_multiplier, std::vector<optimizer_pt> & local_optimizers, const std::vector<solution_pt> & elitist_archive);
    void add_elites_to_archive(std::vector<solution_pt> & elitist_archive, const std::vector<solution_pt> & elite_candidates, int & global_opts_found, int & new_global_opts_found);
    
    // Termination criteria
    //-------------------------------------------------------------------------------
    bool terminate_on_runtime() const;
    bool terminate_on_approaching_elite(optimizer_t & local_optimizer, std::vector<solution_pt> & elite_candidates);
    bool terminate_on_converging_to_local_optimum(optimizer_t & local_optimizer, std::vector<solution_pt> & elite_candidates);
   
    // data members : populations
    //--------------------------------------------------------------------------------
    population_pt pop;	// This refers the sample population

    // Output to file
    //-------------------------------------------------------------------------------
	std::ofstream statistics_file, statistics_file_estimate_time;
    void new_statistics_file(int run_id);
	void new_statistics_file_estimate_time(int run_id);
	void write_statistics_line_population(const population_t& pop, const std::vector<optimizer_pt>& local_optimizers, const std::vector<solution_pt>& elitist_archive);
    void write_statistics_line_cluster(const population_t & cluster_pop,int cluster_number,  int cluster_generation, const std::vector<optimizer_pt> & local_optimizers, const std::vector<solution_pt> & elitist_archive);
	void write_statistics_line_estimate_time(const population_t& pop);
	void close_statistics_file();
	void close_statistics_file_estimate_time();
    void write_population_file(int run_id, population_pt pop, std::vector<optimizer_pt> & local_optimizers) const;
    void write_selection_file(int run_id, population_pt pop, bool initial_mesh, std::vector<optimizer_pt> & local_optimizers) const;
    void write_niche_population(int run_id, bool by_nicher, int generation_number, size_t cluster_number, int cluster_generation, population_pt pop) const;
    void write_elitist_archive_file(int run_id, const std::vector<solution_pt> & elitist_archive) const;
    void write_CEC2013_niching_file(int run_id, bool final);

  };
  
  
  
}
